from django.views.decorators.csrf import csrf_exempt, csrf_protect
from cached_hitcount.settings import CACHED_HITCOUNT_CSRF_EXEMPT

def conditional_csrf_exempt(func):
    if CACHED_HITCOUNT_CSRF_EXEMPT:
        return csrf_exempt(func)
    else:
        return csrf_protect(func)


import datetime
from django.core.management.base import NoArgsCommand
from cached_hitcount.tasks import persist_hits

class Command(NoArgsCommand):
    help = "Can be run as a cronjob or directly to persist_hits to the database."

    def handle_noargs(self, **options):
        persist_hits()



import datetime
from django.core.management.base import NoArgsCommand
from cached_hitcount.utils import release_lock


class Command(NoArgsCommand):
    help = "Will delete the memcache lock used by hitcount"

    def handle_noargs(self, **options):
        release_lock()



from datetime import datetime
from optparse import make_option

from django.core.management.base import BaseCommand

from cached_hitcount.models import Hit
from cached_hitcount.management.helpers import prompt_bool

'''
python manage.py hitcount_clear_hits --from='2014-05-27' --to='2014-05-27'
'''
class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--from',  dest='from_date', action='store', type="string", default=None, help='Datetime from (incl.)'),
        make_option('--to',  dest='to_date', action='store', type="string", default=None, help='Datetime up to (incl.)'),
    )
    help = "Will delete all existing hits"
    def handle(self, **options):
        from_date  = options.get('from_date', None)
        to_date  = options.get('to_date', None)

        hits = Hit.objects.all()

        if from_date is not None:
            parsed_from_date = datetime.strptime( from_date, "%Y-%m-%d" )
            hits = hits.filter(added__gte = parsed_from_date)
        if to_date is not None:
            parsed_to_date = datetime.strptime( to_date, "%Y-%m-%d" )
            hits = hits.filter(added__lte = parsed_to_date)

        print hits.query
        if prompt_bool("Are you sure you want to delete all selected %d hits" % hits.count()):
            hits.delete()
            print "All selected hits have been deleted"





# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Hit'
        db.create_table(u'cached_hitcount_hit', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('hits', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
            ('added', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.utcnow)),
            ('content_type', self.gf('django.db.models.fields.related.ForeignKey')(related_name='content_type_set_for_hit', to=orm['contenttypes.ContentType'])),
            ('object_pk', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True)),
        ))
        db.send_create_signal(u'cached_hitcount', ['Hit'])

        # Adding model 'BlacklistIP'
        db.create_table(u'cached_hitcount_blacklistip', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('ip', self.gf('django.db.models.fields.CharField')(unique=True, max_length=40)),
        ))
        db.send_create_signal(u'cached_hitcount', ['BlacklistIP'])


    def backwards(self, orm):
        # Deleting model 'Hit'
        db.delete_table(u'cached_hitcount_hit')

        # Deleting model 'BlacklistIP'
        db.delete_table(u'cached_hitcount_blacklistip')


    models = {
        u'cached_hitcount.blacklistip': {
            'Meta': {'object_name': 'BlacklistIP'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'ip': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '40'})
        },
        u'cached_hitcount.hit': {
            'Meta': {'ordering': "('-hits',)", 'object_name': 'Hit'},
            'added': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.utcnow'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'content_type_set_for_hit'", 'to': u"orm['contenttypes.ContentType']"}),
            'hits': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'object_pk': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'})
        },
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        }
    }

    complete_apps = ['cached_hitcount']


# -*- coding: utf-8 -*-
import string
import getpass

def to_bool(val, default=False, yes_choices=None, no_choices=None):
    if not isinstance(val, basestring):
        return bool(val)

    yes_choices = yes_choices or ('y', 'yes', '1', 'on', 'true', 't')
    no_choices = no_choices or ('n', 'no', '0', 'off', 'false', 'f')

    if val.lower() in yes_choices:
        return True
    elif val.lower() in no_choices:
        return False
    return default


def prompt(name, default=None):
    """
    Grab user input from command line.

    :param name: prompt text
    :param default: default value if no input provided.
    """

    prompt = name + (' [%s]' % default if default else '')
    prompt += ' ' if name.endswith('?') else ': '
    while True:
        rv = raw_input(prompt)
        if rv:
            return rv
        if default is not None:
            return default


def prompt_pass(name, default=None):
    """
    Grabs hidden (password) input from command line.

    :param name: prompt text
    :param default: default value if no input provided.
    """

    prompt = name + (default and ' [%s]' % default or '')
    prompt += name.endswith('?') and ' ' or ': '
    while True:
        rv = getpass.getpass(prompt)
        if rv:
            return rv
        if default is not None:
            return default


def prompt_bool(name, default=False, yes_choices= ('y', 'yes', '1', 'on', 'true', 't'), no_choices=('n', 'no', '0', 'off', 'false', 'f')):
    """
    Grabs user input from command line and converts to boolean
    value.

    :param name: prompt text
    :param default: default value if no input provided.
    :param yes_choices: default 'y', 'yes', '1', 'on', 'true', 't'
    :param no_choices: default 'n', 'no', '0', 'off', 'false', 'f'
    """

    while True:
        rv = prompt(name + '?', yes_choices[0] if default else no_choices[0])
        if not rv:
            return default
        rv = to_bool(rv, default=None, yes_choices=yes_choices, no_choices=no_choices)
        if rv is not None:
            return rv


def prompt_choices(name, choices, default=None, resolve=string.lower,
                   no_choice=('none',)):
    """
    Grabs user input from command line from set of provided choices.

    :param name: prompt text
    :param choices: list or tuple of available choices. Choices may be
                    single strings or (key, value) tuples.
    :param default: default value if no input provided.
    :param no_choice: acceptable list of strings for "null choice"
    """

    _choices = []
    options = []

    for choice in choices:
        if isinstance(choice, basestring):
            options.append(choice)
        else:
            options.append("[%s] %s" % (choice[0], choice[1]))
            choice = choice[0]
        _choices.append(choice)

    while True:
        rv = prompt(name + '? - %s' % '\n'.join(options), default)
        if not rv:
            return default
        rv = resolve(rv)
        if rv in no_choice:
            return None
        if rv in _choices:
            return rv



import json

from django.http import Http404, HttpResponse
from django.views.decorators.cache import never_cache

from cached_hitcount.utils import get_ip, get_hitcount_cache, is_cached_hitcount_enabled, is_bot_request, using_memcache
from cached_hitcount.models import BlacklistIP
from cached_hitcount.settings import CACHED_HITCOUNT_CACHE_TIMEOUT, CACHED_HITCOUNT_EXCLUDE_IP_ADDRESS, \
    CACHED_HITCOUNT_EXCLUDE_BOTS, CACHED_HITCOUNT_LOCK_KEY, \
    CACHED_HITCOUNT_SERVER_CALLBACKS

from cached_hitcount.decorators import conditional_csrf_exempt

def _update_hit_count(request, object_pk, ctype_pk):
    '''
    Evaluates a request's Hit and corresponding HitCount object and,
    after a bit of clever logic, either ignores the request or registers
    a new Hit.

    This is NOT a view!  But should be used within a view ...

    Returns True if the request was considered a Hit; returns False if not.
    '''
    if request:#we might be calling this form a signal
        ip = get_ip(request)

        # first, check our request against the blacklists before continuing
        if CACHED_HITCOUNT_EXCLUDE_IP_ADDRESS and ip in BlacklistIP.objects.get_cache():
            return False

        #are we excluding bots and is this a bot
        if CACHED_HITCOUNT_EXCLUDE_BOTS and is_bot_request(request):
            return False

    #save to memcache
    hitcount_cache = get_hitcount_cache()

    cache_key = "hitcount__%s__%s" % (ctype_pk, object_pk)

    #if the lock is set then do nothing (this means the hits are being persisted to DB)
    lock = hitcount_cache.get(CACHED_HITCOUNT_LOCK_KEY)
    #print  'check %s lock = %s' % (CACHED_HITCOUNT_LOCK_KEY, lock)
    if lock is None or lock != 1:
        try:
            hitcount_cache.incr(cache_key)
            return True
        except ValueError:#cache might have timed out
            count = 1
            hitcount_cache.set(cache_key, count, CACHED_HITCOUNT_CACHE_TIMEOUT)
            return True

    return False

def json_error_response(error_message):
    return HttpResponse(json.dumps(dict(success=False,
                                              error_message=error_message)))

def call_custom_callbacks():
    results = {}
    #AttributeError
    #ImportError
    for custom_callback_key, custom_callback_module_method in CACHED_HITCOUNT_SERVER_CALLBACKS.iteritems():
        custom_callback_module, custom_callback_method = custom_callback_module_method
        try:
            m = __import__(custom_callback_module, globals(), locals(), [custom_callback_method], -1)
            func = getattr(m,custom_callback_method)
            results[custom_callback_key] = func()
        except AttributeError, ae:
            results.setdefault('errors', []).append('%s has no method %s' % (custom_callback_module, custom_callback_method))
        except ImportError, ie:
            results.setdefault('errors', []).append('Unable to import %s' % custom_callback_module)

    return results

@never_cache
@conditional_csrf_exempt
def update_hit_count_ajax(request):
    '''
    Ajax call that can be used to update a hit count.

    Ajax is not the only way to do this, but probably will cut down on 
    bots and spiders.

    See template tags for how to implement.
    '''

    # make sure this is an ajax request
    if not is_cached_hitcount_enabled() or not using_memcache() or not request.is_ajax():
        raise Http404()

    if request.method == "GET":
        return json_error_response("Hits counted via POST only.")

    object_pk = request.POST.get('object_pk', None)
    ctype_pk = request.POST.get('ctype_pk', None)
    
    status = "no hit recorded"
    if object_pk and ctype_pk:
        result = _update_hit_count(request, object_pk, ctype_pk)

        if result:
            status = "success"

    result_dict = {'status': status}
    if CACHED_HITCOUNT_SERVER_CALLBACKS:
        result_dict.update(call_custom_callbacks())

    return HttpResponse(json.dumps(result_dict),mimetype="application/json")



import datetime

from django.db import models

from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic

from django.dispatch import Signal

from cached_hitcount.settings import CACHED_HITCOUNT_IP_CACHE_TIMEOUT, CACHED_HITCOUNT_IP_CACHE
from cached_hitcount.utils import get_hitcount_cache, get_target_ctype_pk
# SIGNALS #
def process_object_saved(sender, **kwargs):
    """ Called when an object is saved. It will create an initial Hit entry with 0 hits """

    object = kwargs['instance']
    created = kwargs['created']

    if created:
        ctype, object_pk = get_target_ctype_pk(object)
        Hit.objects.select_for_update().get_or_create(added=datetime.datetime.utcnow().date(), object_pk=object_pk, content_type=ctype)
        from cached_hitcount.views import _update_hit_count#prevent circular import
        _update_hit_count(None,object_pk, ctype.pk)

# MODELS #
class Hit(models.Model):
    '''
    Model that stores the hit totals for any content object.

    '''

    hits = models.PositiveIntegerField(default=0)
    added        = models.DateField(default=datetime.datetime.utcnow().date(), db_index=True)
    content_type    = models.ForeignKey(ContentType,
                        verbose_name="content type",
                        related_name="content_type_set_for_%(class)s", db_index=True)
    object_pk       = models.PositiveIntegerField('object ID', db_index=True)
    content_object  = generic.GenericForeignKey('content_type', 'object_pk')

    class Meta:
        ordering = ( '-hits', )
        #unique_together = (("content_type", "object_pk"),)
        get_latest_by = "added"

    def __unicode__(self):
        return u'%s' % self.content_object

    def get_content_object_url(self):
        '''
        Django has this in its contrib.comments.model file -- seems worth
        implementing though it may take a couple steps.
        '''
        pass

#MANAGER
class BlacklistIPManager(models.Manager):
    def set_cache(self):
        hitcount_cache = get_hitcount_cache()
        blacklist_ips = list(self.get_query_set().values_list('ip',flat=True))
        hitcount_cache.set(CACHED_HITCOUNT_IP_CACHE, blacklist_ips, CACHED_HITCOUNT_IP_CACHE_TIMEOUT)
        return blacklist_ips

    def get_cache(self):
        hitcount_cache = get_hitcount_cache()
        blacklist_ips = hitcount_cache.get(CACHED_HITCOUNT_IP_CACHE)
        if not blacklist_ips and not isinstance(blacklist_ips, list):
            blacklist_ips = self.set_cache()
        return blacklist_ips


class BlacklistIP(models.Model):

    objects = BlacklistIPManager()

    ip = models.CharField(max_length=40, unique=True)

    def __unicode__(self):
        return u'%s' % self.ip

    def save(self,*args, **kwargs):
        super(BlacklistIP, self).save(*args, **kwargs)
        BlacklistIP.objects.set_cache()



__author__ = 'philroche'

import unittest
import json
from functools import wraps

from django.utils.decorators import available_attrs
from django.test import Client
from django.core.urlresolvers import reverse

from cached_hitcount.models import Hit
from cached_hitcount.utils import get_target_ctype_pk, get_hit_count, using_memcache, is_cached_hitcount_enabled
from cached_hitcount.tasks import persist_hits


'''
Decorator so that tests are not run if memcache is not in use
'''
def only_if_memcache_in_use(method):
    @wraps(method, assigned=available_attrs(method))
    def wrapper(request, *args, **kwargs):
        if using_memcache():
            return method(request, *args, **kwargs)
        else:
            return True
    return wrapper


class CachedHitCountTestCase(unittest.TestCase):
    def setUp(self):
        self.client = Client()
        #create a model to use to test hits
        self._clear_old_hits()
        self._create_sample_model()


    def _clear_old_hits(self):
        Hit.objects.all().delete()


    def _create_sample_model(self):
        self.sample_model = Hit()
        self.sample_model.object_pk = 435
        ctype, object_pk = get_target_ctype_pk(self.sample_model)
        self.sample_model.content_type = ctype
        self.sample_model.save()

    def test_memcache_in_use(self):
        self.assertTrue(using_memcache(), msg="Memcache is not is use so Django Cached Hitcount will not work")

    def test_enabled(self):
        self.assertTrue(is_cached_hitcount_enabled(), msg="Django Cached Hitcount is not enabled (gargoyle switch)")


    @only_if_memcache_in_use
    def test_can_add_to_cache(self):
        #$.post( '/hits/',{ object_pk : '1', ctype_pk :  '8', csrfmiddlewaretoken: 'fcvMlt1fkwY8qK5NFgazvxy5iaWGGha8'},'json');
        json_string = { 'object_pk' : '%s' % str(self.sample_model.pk), 'ctype_pk' :  '%s' % str(self.sample_model.content_type.pk), 'csrfmiddlewaretoken': 'fcvMlt1fkwY8qK5NFgazvxy5iaWGGha8'}
        response = self.client.post(reverse('update_hit_count_ajax'),
                            json_string,
                            HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        response_data = json.loads(response.content)
        self.assertEqual(response_data["status"],"success", msg="Unable to add hit to hitcount")

    @only_if_memcache_in_use
    def test_can_persist(self):
        self.test_can_add_to_cache()
        try:
            persist_hits()
        except Exception:
            self.fail("persist_hits raised Exception unexpectedly - unable to save hits from cache to DB!")

    @only_if_memcache_in_use
    def test_can_read(self):
        self.test_can_persist()#this will reset the memcached hit count to 0
        self._clear_old_hits()
        self._create_sample_model()
        self.test_can_persist()
        hit_count = get_hit_count(self.sample_model)
        self.assertEqual(hit_count, "1", msg="Unable to read hitcounts from DB")

if __name__ == '__main__':
    unittest.main()


from datetime import datetime, timedelta
from celery.task import periodic_task
from memcached_stats import MemcachedStats
import logging


from django.core.cache import parse_backend_conf
from django.contrib.contenttypes.models import ContentType

try:
    from django.db.transaction import atomic as transaction_atomic
except ImportError:
    from django.db.transaction import commit_on_success as transaction_atomic

from cached_hitcount.utils import get_hitcount_cache, is_cached_hitcount_enabled, release_lock, using_memcache
from cached_hitcount.settings import CACHED_HITCOUNT_CACHE, CACHED_HITCOUNT_CACHE_TIMEOUT, CACHED_HITCOUNT_IP_CACHE, CACHED_HITCOUNT_LOCK_KEY
from cached_hitcount.models import Hit

logger = logging.getLogger(__name__)

@periodic_task(run_every=timedelta(minutes=5))
def persist_hits():
    if is_cached_hitcount_enabled() and using_memcache():

        backend, location, params = parse_backend_conf(CACHED_HITCOUNT_CACHE)
        host, port = location.split(':')
        hitcount_cache = get_hitcount_cache()
        lock = hitcount_cache.get(CACHED_HITCOUNT_LOCK_KEY)
        #print  'persist_hits - check %s lock = %s' % (CACHED_HITCOUNT_LOCK_KEY, lock)
        if lock is None or lock != 1:
            try:
                #acquire a lock so no updates will occur while we are persisting the hits to DB
                hitcount_cache.set(CACHED_HITCOUNT_LOCK_KEY, 1, CACHED_HITCOUNT_CACHE_TIMEOUT)
                #print  'acquire %s lock = %s ' % (CACHED_HITCOUNT_LOCK_KEY, hitcount_cache.get(CACHED_HITCOUNT_LOCK_KEY))
                mem = MemcachedStats(host, port)
                keys = mem.keys()

                content_types = {}#used for keeping track of the content types so DB doesn't have to be queried each time
                for cache_key in keys:
                    if "hitcount__" in cache_key and not CACHED_HITCOUNT_IP_CACHE in cache_key:
                        cache_key = cache_key.split(':')[-1]#the key is a combination of key_prefix, version and key all separated by : - all we need is the key
                        count = hitcount_cache.get(cache_key)
                        if count:#only update the hit count if the is not None
                            hitcount, ctype_pk, object_pk  = cache_key.split('__')
                            if ctype_pk in content_types.keys():
                                content_type = content_types[ctype_pk]
                            else:
                                content_type = ContentType.objects.get(id=ctype_pk)
                                content_types[ctype_pk] = content_type

                            with transaction_atomic():
                                #save a new hit or increment this hits on an existing hit
                                hit, created = Hit.objects.select_for_update().get_or_create(added=datetime.utcnow().date(), object_pk=object_pk, content_type=content_type)
                                if hit and created:
                                    hit.hits = long(count)
                                    hit.save()
                                elif hit:
                                    hit.hits = hit.hits + long(count)
                                    hit.save()

                        #reset the hitcount for this object to 0 - even if it was previously None
                        hitcount_cache.set(cache_key, 0, CACHED_HITCOUNT_CACHE_TIMEOUT)
                        #print  'reset key %s to zero = %s ' % (cache_key, hitcount_cache.get(cache_key))
            except Exception, ex:
                logger.error('Unable to persist hits')
                logger.error(ex)

                raise ex
            finally:
                release_lock()
        else:
            logger.error('Unable to persist hits due to persist task already in progress (%s is locked)' % CACHED_HITCOUNT_LOCK_KEY)








from django.conf import settings

CACHED_HITCOUNT_CACHE = getattr(settings, 'HITCOUNT_CACHE', 'default')#The CACHES settings dict key to use for caching hits # hitcount_cache

CACHED_HITCOUNT_ENABLED = hasattr(settings, 'CACHES') and CACHED_HITCOUNT_CACHE in getattr(settings, 'CACHES', {})

CACHED_HITCOUNT_CACHE_TIMEOUT = getattr(settings, 'HITCOUNT_CACHE_TIMEOUT', 600 )#600 = 10 mins,  3600 = 1 hour# How long in seconds should the hits be cached for before persisted
CACHED_HITCOUNT_IP_CACHE_TIMEOUT = getattr(settings, 'HITCOUNT_IP_CACHE_TIMEOUT', 86400 )#600 = 10 mins,  3600 = 1 hour, 86400=1 day# How long in seconds should the black list ip addresses be cached for
CACHED_HITCOUNT_PERSIST_SCHEDULE = getattr(settings, 'HITCOUNT_PERSIST_SCHEDULE', 15 )#minutes # How often should the hits be persisted
CACHED_HITCOUNT_IP_CACHE = getattr(settings, 'HITCOUNT_IP_CACHE', 'hitcount__blacklistip')#key name to use for caching the blacklist ip addresses
CACHED_HITCOUNT_LOCK_KEY = getattr(settings, 'HITCOUNT_LOCK_KEY', 'hitcountlock')#key name to use for locking when persisting hits to DB
CACHED_HITCOUNT_EXCLUDE_IP_ADDRESS = getattr(settings, 'HITCOUNT_EXCLUDE_IP_ADDRESS', True)#Wheter or not you want to exclude some ip addresses
CACHED_HITCOUNT_EXCLUDE_BOTS = getattr(settings, 'HITCOUNT_EXCLUDE_BOTS', True)#Whether or not you exclude bots
CACHED_HITCOUNT_CSRF_EXEMPT = getattr(settings, 'HITCOUNT_CSRF_EXEMPT', False)#Whether or not ajax view is CSRF exempt by default
CACHED_HITCOUNT_SERVER_CALLBACKS = getattr(settings, 'HITCOUNT_SERVER_CALLBACKS', False)#Dict with with value as tuple of module and method to call on server. (default is {})
CACHED_HITCOUNT_CLIENT_CALLBACKS = getattr(settings, 'HITCOUNT_CLIENT_CALLBACKS', False)#List of Javascript functions to call when response is received. Response is passed to these functions. (default is [])


from django import template
from django.core.urlresolvers import reverse

from cached_hitcount.utils import get_target_ctype_pk, get_hit_count as get_hit_count_utils
from cached_hitcount.settings import CACHED_HITCOUNT_CLIENT_CALLBACKS

register = template.Library()




@register.simple_tag(takes_context=True)
def get_hit_count(context, object, *args, **kwargs):
    '''
    Returns hit counts for an object.

    - Return total hits for an object: 
      {% get_hit_count [object] %}
    
    - Get total hits for an object over a certain time period:
      {% get_hit_count [object] within=["days=1"] %}

    The within arguments need to follow datetime.timedelta's limitations:
    Accepts days and weeks.
    '''
    kwargs['object'] = object
    return get_hit_count_utils(*args, **kwargs)



@register.inclusion_tag('hitcount_js.html', takes_context=True)
def get_hit_count_javascript_template(context, object, **kwargs):
    '''
    Return javascript for an object
    and requires jQuery.  NOTE: only works on a single object, not an object
    list.

    For example:

    {% get_hit_count_javascript_template [object] %}

    '''
    url = reverse('update_hit_count_ajax')
    ctype, object_pk = get_target_ctype_pk(object)
    csrf_token = unicode(context['csrf_token'])

    template_context = {
        'ctype_pk': ctype.pk,
        'object_pk': object_pk,
        'csrf_token': csrf_token,
        'url' : url,
    }

    if CACHED_HITCOUNT_CLIENT_CALLBACKS:
        template_context['client_callbacks'] = CACHED_HITCOUNT_CLIENT_CALLBACKS

    return template_context